# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{udplog.twisted}.
"""

from __future__ import division, absolute_import

import logging

from zope.interface import verify

from twisted.internet import defer
from twisted.internet.interfaces import IPushProducer
from twisted.internet import task
from twisted.python import failure
from twisted.python import log
from twisted.trial import unittest

from udplog import twisted
from udplog import udplog

class UDPLogObserverTest(unittest.TestCase):
    """
    Tests for L{udplog.twisted.UDPLogObserver}.
    """

    def setUp(self):
        self.logger = udplog.MemoryLogger()
        self.observer = twisted.UDPLogObserver(self.logger,
                                               defaultCategory='test')
        self.publisher = log.LogPublisher()
        self.publisher.addObserver(self.observer.emit)


    def test_emit(self):
        """
        Minimal log event is passed to UDP log.
        """
        self.publisher.msg('Test')

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('test', category)
        self.assertEqual('Test', eventDict['message'])
        self.assertIn('timestamp', eventDict)
        self.assertNotIn('time', eventDict)
        self.assertEqual('INFO', eventDict['logLevel'])


    def test_emitCategory(self):
        """
        The value of key 'category' sets the UDP log category for the event.
        """
        self.publisher.msg('Test', category='other')

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('other', category)
        self.assertEqual('Test', eventDict['message'])


    def test_emitErr(self):
        """
        A typical error log event is passed to UDP log.

        This mimics how L{log.err} would call C{msg}.
        """
        self.publisher.msg(failure=failure.Failure(Exception("Test")),
                           why="Something went wrong",
                           isError=1)

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('Traceback (most recent call last):\n'
                         'Failure: exceptions.Exception: Test\n',
                         eventDict['excText'])
        self.assertEqual('exceptions.Exception', eventDict['excType'])
        self.assertEqual('Test', eventDict['excValue'])
        self.assertEqual('Something went wrong', eventDict['message'])
        self.assertNotIn('why', eventDict)
        self.assertNotIn('failure', eventDict)
        self.assertEqual('ERROR', eventDict['logLevel'])


    def test_emitErrNoWhy(self):
        """
        If there is no 'why' log the exception's message.
        """
        self.publisher.msg(failure=failure.Failure(Exception("Test")),
                           why=None,
                           isError=1)

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('Test', eventDict['message'])
        self.assertNotIn('why', eventDict)


    def test_emitErrNoWhyNoArgs(self):
        """
        If there is no 'why' and the exception has no message, log its class.
        """
        self.publisher.msg(failure=failure.Failure(Exception()),
                           why=None,
                           isError=1)

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('exceptions.Exception', eventDict['message'])


    def test_emitWarning(self):
        """
        A typical warning event is passed to UDP log.

        This mimics how warnings are picked up and logged.
        """
        self.publisher.msg(warning=UserWarning("Oops!"),
                           category='exceptions.UserWarning',
                           filename='test.py', lineno=342,
                           format='%(filename)s:%(lineno)s: '
                                  '%(category)s: %(warning)s')

        self.assertEqual(1, len(self.logger.logged))
        category, eventDict = self.logger.logged[-1]

        self.assertEqual('Oops!', eventDict['message'])
        self.assertNotIn('warning', eventDict)
        self.assertEqual('exceptions.UserWarning',
                         eventDict['warningCategory'])
        self.assertEqual('test.py', eventDict['filename'])
        self.assertEqual(342, eventDict['lineno'])
        self.assertEqual('test', category)
        self.assertEqual('WARNING', eventDict['logLevel'])


    def test_start(self):
        """
        Starting the observer adds it to the default log publisher.
        """
        self.observer.start()
        self.assertIn(self.observer.emit, log.theLogPublisher.observers)
        self.observer.stop()


    def test_stop(self):
        """
        Stopping the observer removes it from the default log publisher.
        """
        self.observer = twisted.UDPLogObserver(self.logger)
        self.observer.start()
        self.observer.stop()

        self.assertNotIn(self.observer.emit, log.theLogPublisher.observers)



class TwistedLogHandlerTest(unittest.TestCase):
    """
    Tests for L{udplog.twisted.TwistedLogHandler}.
    """

    def log(self, eventDict):
        self.logged.append(eventDict)


    def setUp(self):
        self.publisher = log.LogPublisher()
        self.publisher.addObserver(self.log)

        self.handler = twisted.TwistedLogHandler(publisher=self.publisher)
        self.logger = logging.Logger('test_logger')
        self.logger.addHandler(self.handler)

        self.logged = []


    def test_emit(self):
        """
        A message logged through python logging ends up in the observer.
        """
        self.logger.debug("Hallo")

        self.assertEqual(1, len(self.logged))
        eventDict = self.logged[-1]

        self.assertEqual('python_logging', eventDict.get('category'))
        self.assertEqual(('Hallo',), eventDict.get('message'))
        self.assertEqual('DEBUG', eventDict.get('logLevel'))
        self.assertEqual('test_logger', eventDict.get('logName'))
        self.assertFalse(eventDict.get('isError'))


    def test_emitException(self):
        """
        Logging an exception captures it in a Failure added to the event.
        """
        try:
            {}['something']
        except Exception:
            self.logger.exception('Oops')

        self.assertEqual(1, len(self.logged))
        eventDict = self.logged[-1]

        self.assertFalse(eventDict.get('message'))
        self.assertEqual('ERROR', eventDict.get('logLevel'))
        self.assertTrue(eventDict.get('isError'))
        self.assertEqual(KeyError, eventDict['failure'].type)
        self.assertEqual('Oops', eventDict['why'])


    def test_emitIsError(self):
        """
        Only messages with log level >= ERROR have isError True.
        """
        def checkIsError(logLevel):
            self.logger.log(logLevel, "Hallo")
            eventDict = self.logged[-1]
            return eventDict['isError']

        self.assertFalse(checkIsError(logging.DEBUG))
        self.assertFalse(checkIsError(logging.INFO))
        self.assertFalse(checkIsError(logging.WARNING))
        self.assertTrue(checkIsError(logging.ERROR))
        self.assertTrue(checkIsError(logging.CRITICAL))



class UDPLogProtocol(unittest.TestCase):
    """
    Tests for L{udplog.twisted.UDPLogProtocol}.
    """

    def setUp(self):
        self.events = []
        self.protocol = twisted.UDPLogProtocol()
        self.protocol.eventReceived = self.events.append


    def test_datagramReceived(self):
        """
        A datagram is a category, a colon and a JSON dict.
        """
        datagram = """test_category:\t{"key": "value"}"""
        self.protocol.datagramReceived(datagram, None)

        self.assertEqual(1, len(self.events))
        event = self.events[-1]

        self.assertEqual('test_category', event.get('category'))
        self.assertEqual('value', event.get('key'))


    def test_datagramReceivedNoMsg(self):
        """
        If there is no colon, a ValueError is logged.
        """
        datagram = """test_category"""
        self.protocol.datagramReceived(datagram, None)
        self.assertEqual(1, len(self.flushLoggedErrors(ValueError)))


    def test_datagramReceivedInvalidJSON(self):
        """
        If the event dictionary is not valid JSON, a ValueError is logged.
        """
        datagram = """test_category:\t{"key":"value"""
        self.protocol.datagramReceived(datagram, None)
        self.assertEqual(1, len(self.flushLoggedErrors(ValueError)))


    def test_datagramReceivedNotDict(self):
        """
        If the encoded JSON is not a dict, a TypeError is logged.
        """
        datagram = """test_category:\t3"""
        self.protocol.datagramReceived(datagram, None)
        self.assertEqual(1, len(self.flushLoggedErrors(TypeError)))



class DispatcherFromUDPLogProtocolTest(unittest.TestCase):
    """
    Tests for L{udplog.twisted.DispatcherFromUDPLogProtocol}.
    """

    def setUp(self):
        self.dispatcher = twisted.DispatcherFromUDPLogProtocol()


    def test_register(self):
        """
        A registered consumer will receive events when dispatched.
        """
        events = []
        self.dispatcher.register(lambda event: events.append(event))
        self.dispatcher.eventReceived(None)
        self.assertEqual([None], events)


    def test_registerMultiple(self):
        """
        Multiple registered consumers will receive events when dispatched.
        """
        events1 = []
        events2 = []
        self.dispatcher.register(lambda event: events1.append(event))
        self.dispatcher.register(lambda event: events2.append(event))
        self.dispatcher.eventReceived(None)
        self.assertEqual([None], events1)
        self.assertEqual([None], events2)


    def test_eventReceivedMultipleException(self):
        """
        A failed consumer does not affect others.
        """
        events = []
        def err(event):
            raise ValueError("Oops")
        self.dispatcher.register(err)
        self.dispatcher.register(lambda event: events.append(event))
        self.dispatcher.eventReceived(None)
        self.assertEqual([None], events)
        self.assertEqual(1, len(self.flushLoggedErrors(ValueError)))


    def test_unregister(self):
        """
        A unregistered consumer will not received events when dispatched.
        """
        events = []
        def consumer(event):
            events.append(event)

        self.dispatcher.register(consumer)
        self.dispatcher.unregister(consumer)
        self.dispatcher.eventReceived(None)
        self.assertEqual([], events)


    def test_unregisterUnknown(self):
        """
        Unknown consumers are silently ignored when unregistered.
        """
        def consumer(event):
            pass

        self.dispatcher.unregister(consumer)



class UDPLogClientFactoryTest(unittest.TestCase):
    """
    Tests for L{UDPLogClientFactory}.
    """

    def setUp(self):
        class TestProtocol(object):
            factory = None

            def __init__(self, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs

        self.factory = twisted.UDPLogClientFactory(TestProtocol, 1, bar=2)


    def test_buildProtocol(self):
        """
        A protocol is created with the args and kwargs passed to the factory.
        """
        protocol = self.factory.buildProtocol(None)
        self.assertEqual((1,), protocol.args)
        self.assertEqual({'bar': 2}, protocol.kwargs)
        self.assertIdentical(self.factory, protocol.factory)


    def test_buildProtocolReset(self):
        """
        When a protocol is built, the factory reconnect delay is reset.
        """
        def protocol():
            pass

        self.factory.delay = self.factory.initialDelay + 5
        self.factory.buildProtocol(None)
        self.assertEqual(self.factory.initialDelay, self.factory.delay)



class QueueProducerTest(unittest.TestCase):
    """
    Tests for L{udplog.twisted.QueueProducer}.
    """

    def setUp(self):
        self.clock = task.Clock()
        self.output = []
        self.producer = twisted.QueueProducer(self.callback, clock=self.clock)


    def callback(self, obj):
        """
        Gets called to deliver an item, saving the item.
        """
        self.output.append(obj)
        return defer.succeed(None)


    def test_interface(self):
        """
        QueueProducer instances provide IPushProducer.
        """
        verify.verifyObject(IPushProducer, self.producer)


    def test_putBeforeResume(self):
        """
        If resumeProducing has never been called, all items are just queued.
        """
        self.producer.put(None)
        self.assertEqual(0, len(self.output))


    def test_putNotPaused(self):
        """
        When not paused has been called, all items are just queued.
        """
        self.producer.resumeProducing()
        self.producer.put(None)
        self.assertEqual(1, len(self.output))


    def test_processQueueReschedule(self):
        """
        After an item has been delivered, subsequent items should, too.
        """
        self.producer.resumeProducing()
        self.producer.put(None)
        self.assertEqual(1, len(self.output))

        self.producer.put(None)
        self.clock.advance(0)
        self.assertEqual(2, len(self.output))

        self.producer.put(None)
        self.producer.put(None)
        self.clock.advance(0)
        self.assertEqual(4, len(self.output))


    def test_processQueueRescheduleError(self):
        """
        Rescheduling happens if the callback returns a failure.
        """
        def callback(obj):
            if obj == 1:
                return defer.fail(ValueError())
            if obj == 2:
                self.output.append(obj)
                return defer.succeed(None)

        self.producer = twisted.QueueProducer(callback, clock=self.clock)
        self.producer.resumeProducing()
        self.producer.put(1)
        self.producer.put(2)
        self.clock.advance(0)
        self.assertEqual([2], self.output)
        self.assertEqual(1, len(self.flushLoggedErrors(ValueError)))



    def test_processQueueRescheduleRealReactor(self):
        """
        Rescheduling happens with the real reactor.
        """
        d = defer.Deferred()
        def callback(obj):
            if obj == 1:
                self.output.append(obj)
            if obj == 2:
                d.callback(obj)
            return defer.succeed(None)

        self.producer = twisted.QueueProducer(callback)
        self.producer.resumeProducing()
        self.producer.put(1)
        self.assertEqual(1, len(self.output))

        self.producer.put(2)
        return d


    def test_pauseProducing(self):
        """
        Items put while paused, are not delivered.
        """
        self.producer.resumeProducing()
        self.producer.pauseProducing()
        self.producer.put(None)
        self.clock.advance(0)
        self.assertEqual(0, len(self.output))


    def test_resumeProducing(self):
        """
        Items put while paused, are delivered after resuming.
        """
        self.producer.resumeProducing()
        self.producer.pauseProducing()
        self.producer.put(None)
        self.clock.advance(0)
        self.producer.resumeProducing()
        self.assertEqual(1, len(self.output))


    def test_stopProducingCancelReschedule(self):
        """
        Stop producing cancels reschedules.
        """
        self.producer.resumeProducing()
        self.producer.put(None)
        self.producer.pauseProducing()
        self.producer.put(None)
        self.producer.stopProducing()
        self.clock.advance(0)
        self.assertEqual(1, len(self.output))


    def test_stopProducingCancelDeferred(self):
        """
        Stop producing cancels the deferred to process the next item.
        """
        self.producer.resumeProducing()
        self.producer.stopProducing()
        self.producer.put(None)
        self.clock.advance(0)
        self.assertEqual(0, len(self.output))


    def test_processQueueRescheduleDeferred(self):
        """
        If the callback deferred fires after pausing, don't deliver.
        """
        d = defer.Deferred()

        def callback(obj):
            if obj == 1:
                return defer.succeed(None)
            elif obj == 2:
                return d
            else:
                self.output.append(obj)
                return defer.succeed(None)

        self.producer = twisted.QueueProducer(callback, clock=self.clock)
        self.producer.resumeProducing()
        self.producer.put(1)
        self.clock.advance(0)
        self.producer.put(2)
        self.producer.put(3)
        self.clock.advance(0)
        self.producer.pauseProducing()
        d.callback(None)
        self.clock.advance(0)
        self.producer.put(3)
        self.clock.advance(0)
        self.assertEqual(0, len(self.output))


    def test_processQueueResumeDeferredReuse(self):
        """
        Resuming right after pausing reuses the existing waiting deferred.
        """
        self.producer.resumeProducing()
        self.producer.put(1)
        self.clock.advance(0)
        self.producer.pauseProducing()
        d = self.producer.waiting
        self.producer.resumeProducing()
        self.clock.advance(0)
        self.assertIdentical(d, self.producer.waiting)


    def test_putLimitedQueue(self):
        """
        Putting items over the limit while paused keeps last n items.
        """
        self.producer = twisted.QueueProducer(self.callback, size=3,
                                             clock=self.clock)
        self.producer.resumeProducing()
        self.producer.pauseProducing()
        self.producer.put(1)
        self.producer.put(2)
        self.producer.put(3)
        self.producer.put(4)
        self.producer.resumeProducing()
        self.clock.advance(0)
        self.assertEqual([2, 3, 4], self.output)






# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{udplog.udplog}.
"""

from __future__ import division, absolute_import

import errno
import logging
import socket
import sys
import time

import StringIO

from twisted.trial import unittest

from udplog import udplog

class UDPLoggerTest(unittest.TestCase):
    """
    Tests for {udplog.udplog.UDPLogger}.
    """

    MAX_DATAGRAM_SIZE = 8192

    def setUp(self):
        self.output = []


    def send(self, data):
        """
        Fake socket.send that records all log events.
        """
        if len(data) > self.MAX_DATAGRAM_SIZE:
            raise socket.error(errno.EMSGSIZE, "Message too long")
        else:
            self.output.append(data)


    def _catchOutput(self, logger):
        self.patch(logger.socket, "send", self.send)
        return logger


    def test_log(self):
        """
        A log event is serialized and sent out over UDP.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        logger.log('test', {u'message': u'test'})

        self.assertEqual(1, len(self.output))

        msg = self.output[0]
        self.assertRegexpMatches(msg, r'^test:\t{.*}$')

        category, eventDict = udplog.unserialize(msg)

        self.assertEqual(b'test', category)
        self.assertIn(u'message', eventDict)
        self.assertEqual(u'test', eventDict[u'message'])


    def test_logNotDict(self):
        """
        If eventDict is not a dict, TypeError is raised.
        """
        logger = udplog.UDPLogger()
        self.assertRaises(AttributeError, logger.log, 'acategory', 1)


    def test_logObjects(self):
        """
        Arbitrary objects do not choke the JSON encoding.
        """
        class Something(object):
            pass

        something = Something()

        logger = udplog.UDPLogger()
        self._catchOutput(logger)
        logger.log('atest', {u'something': something})

        category, eventDict = udplog.unserialize(self.output[0])
        self.assertEqual(repr(something), eventDict[u'something'])


    def test_logNonUnicode(self):
        """
        Non-utf8-encodable dicts raise a UnicodeDecodeError.

        Ensure passing an *utf8 encodable* dict to udplog, otherwise you will
        make simplejson cry.
        """
        logger = udplog.UDPLogger()
        self.assertRaises(UnicodeDecodeError, logger.log,
                          'atest', {u'good': u'abc', u'bad': b'\x80abc' })


    def test_logTooLong(self):
        """
        If the log event is too long to fit in a UDP datagram, send regrets.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        logger.log('atest', {u'message': u'a' * self.MAX_DATAGRAM_SIZE,
                             u'timestamp': 1357328823.75116})

        self.assertEqual(1, len(self.output))
        category, eventDict = udplog.unserialize(self.output[0])

        self.assertEqual('udplog', category)
        self.assertEqual(u'Failed to send udplog message',
                         eventDict[u'message'])
        self.assertEqual(u'socket.error', eventDict['excType'])
        self.assertEqual(u'[Errno %d] Message too long' % errno.EMSGSIZE,
                         eventDict[u'excValue'])
        self.assertIn(u'excText', eventDict)
        self.assertEqual(u'WARNING', eventDict[u'logLevel'])

        # Check trimmed original
        original = eventDict[u'original']
        self.assertEqual(u'atest', original[u'category'])
        self.assertEqual(1357328823.75116, original[u'timestamp'])
        self.assertEqual(u'a' * (udplog.MAX_TRIMMED_MESSAGE_SIZE - 4) + '[..]',
                         original[u'message'])

        self.assertEqual(self.MAX_DATAGRAM_SIZE,
                         eventDict[u'original_message_size'])
        self.assertLess(self.MAX_DATAGRAM_SIZE,
                        eventDict[u'original_size'])


    def test_logTooLongAdditionalFields(self):
        """
        If the log event is too long, keep several fields.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        eventDict = {
            u'message': u'a' * self.MAX_DATAGRAM_SIZE,
            u'timestamp': 1357328823.75116,
            u'logLevel': u'ERROR',
            u'logName': __name__,
            u'excType': u'exceptions.ValueError',
            u'excValue': 'Oops',
            u'excText': u'exceptions.ValueError: Oops',
            u'filename': __file__,
            u'lineno': 4,
            u'funcName': u'test_log_too_long_additional_fields',
            u'foo': u'bar',
            }

        logger.log('atest', eventDict)
        self.assertEqual(1, len(self.output))

        category, failEventDict = udplog.unserialize(self.output[0])
        original = failEventDict[u'original']

        self.assertEqual(eventDict[u'logLevel'], original[u'logLevel'])
        self.assertEqual(eventDict[u'logName'], original[u'logName'])
        self.assertEqual(eventDict[u'excType'], original[u'excType'])
        self.assertEqual(eventDict[u'excValue'], original[u'excValue'])
        self.assertEqual(eventDict[u'excText'], original[u'excText'])
        self.assertEqual(eventDict[u'filename'], original[u'filename'])
        self.assertEqual(eventDict[u'lineno'], original[u'lineno'])
        self.assertEqual(eventDict[u'funcName'], original[u'funcName'])
        self.assertNotIn(u'foo', original)


    def test_logTooLongCategory(self):
        """
        If the log category is way too long, an exception is printed to stderr.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        eventDict = {
            u'message': u'a' * self.MAX_DATAGRAM_SIZE,
            u'timestamp': 1357328823.75116,
            }

        self.addCleanup(setattr, sys, 'stderr', sys.stderr)
        sys.stderr = StringIO.StringIO()

        logger.log(u'a' * self.MAX_DATAGRAM_SIZE, eventDict)
        self.assertEqual(0, len(self.output))
        self.assertRegexpMatches(
            sys.stderr.getvalue(),
            r'^Failed to send udplog message\n.*Message too long')


    def test_logTooLongTimestamp(self):
        """
        If the log event is too long, the warning has a timestamp.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        logger.log('atest', {u'message': u'a' * self.MAX_DATAGRAM_SIZE,
                             u'timestamp': 1357328823.75116})

        self.assertEqual(1, len(self.output))
        category, eventDict = udplog.unserialize(self.output[0])

        self.assertIn('timestamp', eventDict)


    def test_logTooLongDefault(self):
        """
        If the log event is too long, the warning has default fields.
        """
        logger = udplog.UDPLogger(defaultFields={'foo': 'bar'})
        self._catchOutput(logger)

        logger.log('atest', {u'message': u'a' * self.MAX_DATAGRAM_SIZE,
                             u'timestamp': 1357328823.75116})

        self.assertEqual(1, len(self.output))
        category, eventDict = udplog.unserialize(self.output[0])

        self.assertIn('foo', eventDict)


    def test_augmentTimestamp(self):
        """
        Every log event gets a timestamp if not already set.
        """
        logger = udplog.UDPLogger()
        self._catchOutput(logger)

        before = time.time()
        logger.log('test', {u'message': u'test'})
        after = time.time()

        category, eventDict = udplog.unserialize(self.output[0])

        self.assertIn(u'timestamp', eventDict)

        timestamp = eventDict[u'timestamp']
        self.assertGreaterEqual(timestamp, before)
        self.assertLessEqual(timestamp, after)


    def test_augmentDefaultFields(self):
        """
        Every log event gets default fields.
        """
        defaultFields = {u'hostname': u'foo.example.org'}

        logger = udplog.UDPLogger(defaultFields=defaultFields)
        self._catchOutput(logger)

        logger.log('test', {u'message': u'test'})

        category, eventDict = udplog.unserialize(self.output[0])

        self.assertIn(u'hostname', eventDict)
        self.assertEqual(u'foo.example.org', eventDict[u'hostname'])


    def test_augmentDefaultFieldsOverride(self):
        """
        Default fields can be overridden in individual events.
        """
        defaultFields = {u'hostname': u'foo.example.org'}

        logger = udplog.UDPLogger(defaultFields=defaultFields)
        self._catchOutput(logger)

        logger.log('test', {u'message': u'test',
                            u'hostname': u'bar.example.org'})

        category, eventDict = udplog.unserialize(self.output[0])

        self.assertIn(u'hostname', eventDict)
        self.assertEqual(u'bar.example.org', eventDict[u'hostname'])



class UDPLogHandlerTest(unittest.TestCase):
    """
    Tests for L{udplog.logging.UDPLogHandler}.
    """

    def setUp(self):
        self.udplogger = udplog.MemoryLogger()
        self.handler = udplog.UDPLogHandler(self.udplogger, category='test')
        self.logger = logging.Logger('test_logger')
        self.logger.addHandler(self.handler)


    def test_emit(self):
        """
        A message logged through python logging is sent out over udp.
        """
        self.logger.debug("Hello")

        self.assertEqual(1, len(self.udplogger.logged))
        category, eventDict = self.udplogger.logged[-1]

        self.assertEqual('Hello', eventDict.get('message'))
        self.assertEqual('DEBUG', eventDict.get('logLevel'))
        self.assertEqual('test_logger', eventDict.get('logName'))
        self.assertIn('timestamp', eventDict)


    def test_emitFormatted(self):
        """
        Messages are formatted and arguments are included in the event.
        """
        self.logger.debug("Hello, %(object)s!", {'object': "world"})

        self.assertEqual(1, len(self.udplogger.logged))
        category, eventDict = self.udplogger.logged[-1]

        self.assertEqual('Hello, world!', eventDict.get('message'))
        self.assertEqual('world', eventDict.get('object'))


    def test_emitException(self):
        """
        Logging an exception renders the traceback.
        """
        try:
            {}['something']
        except Exception:
            self.logger.exception('Oops')

        self.assertEqual(1, len(self.udplogger.logged))
        _, eventDict = self.udplogger.logged[-1]

        self.assertEqual('Oops', eventDict.get('message'))
        self.assertEqual('ERROR', eventDict.get('logLevel'))
        self.assertEqual('exceptions.KeyError', eventDict.get('excType'))
        self.assertEqual("'something'", eventDict.get('excValue'))
        self.assertTrue(eventDict.get('excText').startswith('Traceback'))


    def test_emitExceptionWithoutContext(self):
        """
        Logging an exception without context succeeds.

        If L{logging.Logger.exception} is called outside of an C{except} block,
        exception context might be lost. The result of the internal call to
        C{sys.exc_info} then yields a tuple of three C{None}s, which should
        simply result in the C{excValue} being C{'None'}.
        """
        self.logger.exception('Oops')

        self.assertEqual(1, len(self.udplogger.logged))
        _, eventDict = self.udplogger.logged[-1]

        self.assertEqual('Oops', eventDict.get('message'))
        self.assertEqual('ERROR', eventDict.get('logLevel'))
        self.assertIdentical('NoneType', eventDict.get('excType'))
        self.assertEqual('None', eventDict.get('excValue'))
        self.assertIdentical(None, eventDict['excText'])


    def test_emit_extra(self):
        """
        Values passed in the extra keyword argument are added to the eventDict.
        """
        self.logger.debug("Hello", extra={'foo': 'bar'})

        category, eventDict = self.udplogger.logged[-1]
        self.assertIn('foo', eventDict)
        self.assertEqual('bar', eventDict['foo'])



class UDPLogHandlerFactoryTest(unittest.TestCase):
    """
    Tests for L{udplog.ConfigurableUDPLogHandler}.
    """

    def test_argsDefaults(self):
        """
        Without arguments, the logger and handler have their defaults.
        """
        handler = udplog.ConfigurableUDPLogHandler()
        logger = handler.logger

        self.assertEquals('python_logging', handler.category)
        self.assertEquals(('127.0.0.1', 55647), logger.socket.getpeername())
        self.assertEquals({'hostname': socket.gethostname()},
                          logger.defaultFields)


    def test_args(self):
        """
        All arguments are passed on.
        """
        handler = udplog.ConfigurableUDPLogHandler({'foo': 'bar'}, 'test',
                                                   '10.0.0.1', 55648, False)
        logger = handler.logger

        self.assertEquals('test', handler.category)
        self.assertEquals(('10.0.0.1', 55648), logger.socket.getpeername())
        self.assertNotIn('hostname', logger.defaultFields)
        self.assertEquals('bar', logger.defaultFields['foo'])



# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{udplog.rabbitmq}.
"""

from __future__ import division, absolute_import

import simplejson

from twisted.internet import defer
from twisted.internet import error
from twisted.internet import task
from twisted.trial import unittest

from txamqp.protocol import AMQClient

from udplog import rabbitmq
from udplog import twisted


class FakeAMQChannel(object):
    """
    Fake AMQ channel that logs all calls to C{basic_publish}.
    """
    def __init__(self):
        self.published = []


    def basic_publish(self, **kwargs):
        self.published.append(kwargs)
        return defer.Deferred()



class RabbitMQPublisherTest(unittest.TestCase):
    """
    Tests for L{udplog.rabbitmq.RabbitMQPublisherTest}.
    """

    def setUp(self):
        self.publisher = rabbitmq.RabbitMQPublisher(None)


    def test_sendEvent(self):
        """
        An event is encoded to JSON and then sent to the channel.
        """
        self.publisher.chan = FakeAMQChannel()
        event = {'message': 'test',
                 'timestamp': 1340634165}
        self.publisher.sendEvent(event)

        output = self.publisher.chan.published[-1]
        body = output['content'].body
        eventDict = simplejson.loads(body)

        self.assertEqual("test", eventDict['message'])


    def test_sendEventTimestampString(self):
        """
        The event timestamp is converted to a string before sending it.
        """
        self.publisher.chan = FakeAMQChannel()
        event = {'message': 'test',
                 'timestamp': 1340634165}
        self.publisher.sendEvent(event)

        output = self.publisher.chan.published[-1]
        body = output['content'].body
        eventDict = simplejson.loads(body)

        self.assertEqual("1340634165", eventDict['timestamp'])
        self.assertEqual(1340634165, event['timestamp'])


    def test_sendEventIsErrorBoolean(self):
        """
        If 'isError' is present, it is converted to a boolean.
        """
        self.publisher.chan = FakeAMQChannel()
        event = {'message': 'test',
                 'timestamp': 1340634165,
                 'isError': 1}
        self.publisher.sendEvent(event)

        output = self.publisher.chan.published[-1]
        body = output['content'].body
        eventDict = simplejson.loads(body)

        self.assertIsInstance(eventDict['isError'], bool)
        self.assertTrue(eventDict['isError'])


    def test_connectionLost(self):
        """
        The producer is unregistered at the dispatcher on lost connection.
        """
        connectionLostCalls = []
        callbackCalls = []

        def connectionLost(self, reason):
            connectionLostCalls.append(reason)

        def cb(event):
            raise Exception("Unexpected callback")

        # Set up publisher as if connected
        clock = task.Clock()
        self.publisher.producer = twisted.QueueProducer(callback=cb,
                                                       clock=clock)
        self.publisher.producer.resumeProducing()
        self.publisher.dispatcher = twisted.DispatcherFromUDPLogProtocol()
        self.publisher.dispatcher.register(self.publisher.producer.put)

        # Patch parent class to test up-call
        self.patch(AMQClient, 'connectionLost', connectionLost)

        # Drop connection
        self.publisher.connectionLost(error.ConnectionDone())
        self.assertEqual(1, len(connectionLostCalls))
        self.assertEqual(1, len(self.flushLoggedErrors(error.ConnectionDone)))

        # Test that the producer was unregistered with the dispatcher
        event = {'message': 'test'}
        self.publisher.dispatcher.eventReceived(event)
        clock.advance(0)
        self.assertEqual(0, len(callbackCalls))






# -*- test-case-name: udplog.test.test_udplog -*-
#
# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Send and receive structured log events over UDP.

This module provides L{UDPLogger} for sending out structured log events. When
used as a script it listens for log events and prints them to standard out.

While this module uses a few helper functions from Twisted, it does not depend
on the Twisted reactor and can be used in any Python project.
"""

from __future__ import division, absolute_import

import logging
import socket
import time

import simplejson

from twisted.python import reflect
from twisted.python.failure import Failure

MAX_TRIMMED_MESSAGE_SIZE = 200

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 55647

# As LogRecord instances will get elements of the 'extra' keyword argument
# to the logging methods bolted on it as attributes, we need to know which
# fields are the non-extra ones. The following creates an empty LogRecord
# to list the fields it sets by default.
__emptyLogRecord = logging.LogRecord(None, None, None, None, None, None, None)
_DEFAULT_LOGGING_ATTRIBUTES = ['message', 'asctime']
_DEFAULT_LOGGING_ATTRIBUTES.extend(vars(__emptyLogRecord).keys())
del __emptyLogRecord

def unserialize(msg):
    """
    Unserialize a log event.

    A log event is defined as a category, followed by a colon, optional
    whitespace and a event dictionary serialized as JSON.

    @return: The category and event dictionary.
    @rtype: C{tuple} of (C{unicode} and C{dict}.
    """

    category, data = msg.split(':', 1)
    return category, simplejson.loads(data)



def augmentWithFailure(eventDict, failure, why=None):
    """
    Augment a log event with exception information.
    """
    eventDict['excText'] = failure.getTraceback()
    eventDict['excType'] = reflect.qual(failure.type)
    eventDict['excValue'] = reflect.safe_str(failure.value)
    eventDict.setdefault('logLevel', 'ERROR')

    eventDict['message'] = (why or
                            eventDict['excValue'] or
                            eventDict['excType'])



class MemoryLogger(object):
    """
    Keeper of all logs in memory.
    """

    def __init__(self):
        self.logged = []


    def log(self, category, eventDict):
        self.logged.append((category, eventDict))



class UDPLogger(object):
    """
    Dispatcher of structured log events over UDP.
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT,
                       defaultFields=None):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.connect((host, port))

        self.defaultFields = defaultFields or {}


    def augment(self, eventDict):
        """
        Augment the event dictionary with timestamp and default fields.
        """
        eventDict.setdefault('timestamp', time.time())
        for key, value in self.defaultFields.iteritems():
            eventDict.setdefault(key, value)


    def serialize(self, category, eventDict):
        """
        Serialize a log event.

        The dictionary is serialized to JSON. To minimize serialization
        failures, for unserializable objects it falls back to the L{repr} of
        such objects.

        @type category: C{str}.
        @type eventDict: L{dict}.
        """
        msg = simplejson.dumps(eventDict,
                               default=lambda x: str(repr(x)),
                               skipkeys=True)
        return "%s:\t%s" % (category, msg)


    def serializeFailure(self, category, eventDict, size, failure, why):
        """
        Serialize a log event for a failed attempt to send a udplog event.

        If present, this truncates the C{'message'} field to
        C{MAX_TRIMMED_MESSAGE_SIZE}, and preserves exception-related fields
        from the original C{eventDict} as C{'original'}.
        """
        newEventDict = {
            'logLevel': 'WARNING',
            'original': {
                'category': category,
                'timestamp': eventDict['timestamp'],
                },
            }

        augmentWithFailure(newEventDict, failure, why)

        # include (trimmed) original message text
        if 'message' in eventDict:
            text = eventDict['message']
            if len(text) > MAX_TRIMMED_MESSAGE_SIZE:
                newEventDict['original_message_size'] = len(text)
                text = text[:MAX_TRIMMED_MESSAGE_SIZE-4] + '[..]'
            newEventDict['original']['message'] = text

        for key in ('logLevel', 'logName', 'excText', 'excType', 'excValue',
                    'lineno', 'filename', 'funcName'):
            if key in eventDict:
                newEventDict['original'][key] = eventDict[key]

        newEventDict['original_size'] = size

        self.augment(newEventDict)
        return self.serialize('udplog', newEventDict)


    def log(self, category, eventDict):
        """
        Log an event.

        @param category: A short string identifying the type of log event.
            The receiving log server may use this to collect all messages of the
            same category in their own log files.
        @type category: C{bytes}

        @param eventDict: The event dictionary. As this is serialized to JSON
            (see L{serialize}), for complex values, you may want to render them
            to a string before adding them to the event dictionary.
        @type eventDict: C{dict}
        """
        self.augment(eventDict)
        data = self.serialize(category, eventDict)

        try:
            self.socket.send(data)
        except:
            failure = Failure()
            why = "Failed to send udplog message"
            data = self.serializeFailure(category, eventDict, len(data),
                                         failure, why)
            try:
                self.socket.send(data)
            except Exception:
                import sys
                text = why + '\n' + failure.getBriefTraceback()
                print >> sys.stderr, text



class UDPLogHandler(logging.Handler):
    """
    Python Logging handler that emits to UDP.
    """

    def __init__(self, logger, category='python_logging'):
        """
        @type logger: L{UDPLogger}.
        """
        logging.Handler.__init__(self)
        self.setFormatter(logging.Formatter())

        self.logger = logger
        self.category = category


    def emit(self, record):
        """
        Emit a record.
        """
        try:
            eventDict = {
                    'category': self.category,
                    'logLevel': record.levelname,
                    'logName': record.name,
                    'filename': record.pathname,
                    'lineno': record.lineno,
                    'funcName': record.funcName,
                    'timestamp': record.created,
                    }

            if isinstance(record.args, dict):
                eventDict.update(record.args)

            extra = {name: value for name, value in vars(record).iteritems()
                     if name not in _DEFAULT_LOGGING_ATTRIBUTES}

            eventDict.update(extra)

            # Format the message for its side effects and extract the message
            # and exception information
            self.format(record)
            eventDict['message'] = record.message
            if record.exc_info:
                excType, excValue = record.exc_info[0:2]
                eventDict['excValue'] = reflect.safe_str(excValue)
                if excValue is None:
                    eventDict['excText'] = None
                    eventDict['excType'] = 'NoneType'
                else:
                    eventDict['excText'] = record.exc_text
                    eventDict['excType'] = reflect.qual(excType)

            # Extract the category, possibly overridden from record.args.
            category =  eventDict['category']
            del eventDict['category']


            self.logger.log(category, eventDict)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)



class ConfigurableUDPLogHandler(UDPLogHandler):
    """
    Configurable UDPLog logging handler.

    This is a convenience subclass of UDPLogHandler for use in logging
    configuration files (see L{logging.config.fileConfig})::

        [loggers]
        keys = root

        [handlers]
        keys = udplog

        [formatters]
        keys =

        [logger_root]
        level = INFO
        handlers = udplog

        [handler_udplog]
        class = udplog.udplog.ConfigurableUDPLogHandler
        level = INFO
        args = ({'appname': 'example'},)

    @note: This is a subclass instead of a factory function because
        L{logging.config} requires this.
    """


    def __init__(self, defaultFields=None, category='python_logging',
                       host=DEFAULT_HOST, port=DEFAULT_PORT,
                       includeHostname=True):
        """
        Set up a UDPLogHandler with a UDPLogger.

        @param defaultFields: Mapping of default fields to include in all
            events.
        @type defaultFields: L{dict}.

        @param category: The UDPLog category.
        @type category: L{bytes}.

        @param host: The UDP host to send to.
        @type host: L{bytes}.

        @param port: The UDP port to send to.
        @type host: L{int}.

        @param includeHostname: If set, the default fields include a
            C{'hostname'} field set to the current hostname.
        @type includeHostname: L{bool}.
        """
        defaultFields = defaultFields or {}

        if includeHostname:
            defaultFields.setdefault('hostname', socket.gethostname())

        logger = UDPLogger(host=host, port=port, defaultFields=defaultFields)
        UDPLogHandler.__init__(self, logger, category)



def main():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((DEFAULT_HOST, DEFAULT_PORT))
    while True:
        data, addr = sock.recvfrom(65535)
        print data.rstrip()



if __name__ == '__main__':
    main()



# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Asynchronous Scribe client support.

This provides a Twisted based Scribe client with an asynchronous interface for
sending logs and a consumer for L{udplog.twisted.DispatcherFromUDPLogProtocol}.
"""

from __future__ import division, absolute_import

import copy
import logging

import simplejson

from twisted.internet import defer
from twisted.internet import protocol
from twisted.python import log

from scribe import scribe
from thrift.Thrift import TApplicationException, TMessageType
from thrift.protocol import TBinaryProtocol
from thrift.transport import TTwisted

class AsyncScribeClient(scribe.Client):
    """
    Asynchronous Scribe client.

    This derives from L{scribe.Client} to work with the Twisted Thrift
    transport and provide an asynchronous interface for L{Log}.

    @ivar _reqs: List of pending requests. When a result comes in, the
    associated deferred will be fired. If the connection is closed,
    the deferreds of the pending requests will be fired with an exception.
    """


    def __init__(self, transport, factory):
        """
        Set up a scribe client.

        @param transport: The transport of the connection to the Scribe
            server.

        @param factory: The protocol factory of the Thrift transport
            protocol.
        """
        scribe.Client.__init__(self, factory.getProtocol(transport))
        self._reqs = {}


    def Log(self, messages):
        """
        Log messages.

        @param messages: The messages to be sent.
        @type messages: C{list} of L{scribe.LogEntry}.

        @return: L{Deferred<twisted.internet.defer.Deferred>}.
        """
        d = defer.Deferred()
        self._reqs[self._seqid] = d
        self.send_Log(messages)
        return d


    def send_Log(self, messages):
        """
        Called to send log messages.
        """
        scribe.Client.send_Log(self, messages)
        self._seqid += 1


    def recv_Log(self, iprot, mtype, rseqid):
        """
        Called when the result of the log request was received.
        """
        if mtype == TMessageType.EXCEPTION:
            result = TApplicationException()
        else:
            result = scribe.Log_result()

        result.read(iprot)
        iprot.readMessageEnd()

        try:
            d = self._reqs.pop(rseqid)
        except KeyError:
            log.err(result, "Unexpected log result")

        if isinstance(result, Exception):
            d.errback(result)
        elif result.success is not None:
            d.callback(result.success)
        else:
            d.errback(TApplicationException(
                TApplicationException.MISSING_RESULT,
                'Log failed: unknown result'))



class ScribeProtocol(TTwisted.ThriftClientProtocol):
    """
    Scribe protocol.

    This connects an asynchronous Scribe client to a server and sends
    out log events from C{dispatcher}.
    """

    def __init__(self, dispatcher, minLogLevel=logging.INFO):
        self.dispatcher = dispatcher
        self.minLogLevel = minLogLevel

        factory = TBinaryProtocol.TBinaryProtocolFactory(strictRead=False,
                                                         strictWrite=False)
        TTwisted.ThriftClientProtocol.__init__(self, AsyncScribeClient,
                                                     factory)


    def connectionMade(self):
        """
        Add this protocol as a consumer of log events.
        """
        TTwisted.ThriftClientProtocol.connectionMade(self)
        self.dispatcher.register(self.sendEvent)


    def connectionLost(self, reason=protocol.connectionDone):
        """
        Remove this protocol as a consumer of log events.
        """
        self.dispatcher.unregister(self.sendEvent)
        TTwisted.ThriftClientProtocol.connectionLost(self, reason)


    def sendEvent(self, event):
        """
        Write an event to Scribe.
        """
        event = copy.copy(event)

        # Drop events with a log level lower than the configured minimum.
        logLevel = logging.getLevelName(event.get('logLevel', 'INFO'))
        if logLevel < self.minLogLevel:
            return

        category = event['category']
        del event['category']

        try:
            message = simplejson.dumps(event)
        except ValueError, e:
            log.err(e, "Could not encode event to JSON")
            return

        entry = scribe.LogEntry(category=category, message=message)
        d = self.client.Log(messages=[entry])
        d.addErrback(log.err)



# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Twisted Application set up for UDPLog.
"""

from __future__ import division, absolute_import

from twisted.application import service
from twisted.application import internet
from twisted.python import usage

from udplog.twisted import DispatcherFromUDPLogProtocol
from udplog.twisted import UDPLogClientFactory
from udplog.twisted import UDPLogToTwistedLog
from udplog import udplog

class Options(usage.Options):
    optParameters = [
        ('udplog-interface', None, udplog.DEFAULT_HOST, 'UDPLog interface'),
        ('udplog-port', None, udplog.DEFAULT_PORT, 'UDPLog port'),

        ('scribe-host', None, None, 'Scribe Thrift host'),
        ('scribe-port', None, 1463, 'Scribe Thrift port'),

        ('rabbitmq-host', None, None, 'RabbitMQ host'),
        ('rabbitmq-port', None, 5672, 'RabbitMQ host'),
        ('rabbitmq-vhost', None, '/', 'RabbitMQ virtual host'),
        ('rabbitmq-exchange', None, 'logs', 'RabbitMQ exchange'),
        ('rabbitmq-queue-size', None, 2500,
             'Maximum number of log events to buffer for RabbitMQ'),
        ('redis-port', None, 6379, 'Redis port'),
        ('redis-key', None, None, 'Redis list key'),
        ]

    optFlags = [
        ('verbose', 'v', 'Log all incoming messages')
        ]

    def __init__(self):
        super(Options, self).__init__()
        self['redis-hosts'] = set()


    def opt_redis_host(self, host):
        """
        Redis host. Repeat for round-robin dispatching.
        """
        self['redis-hosts'].add(host)



def makeService(config):

    s = service.MultiService()

    # Set up UDP server as the dispatcher.
    dispatcher = DispatcherFromUDPLogProtocol()

    udplogServer = internet.UDPServer(port=config['udplog-port'],
                                      protocol=dispatcher,
                                      interface=config['udplog-interface'],
                                      maxPacketSize=65536)
    udplogServer.setServiceParent(s)

    # Set up Thrift/Scribe client.
    if config['scribe-host']:
        from udplog import scribe
        factory = UDPLogClientFactory(scribe.ScribeProtocol,
                                      dispatcher)
        scribeClient = internet.TCPClient(config['scribe-host'],
                                          config['scribe-port'],
                                          factory)
        scribeClient.setServiceParent(s)

    # Set up RabbitMQ client.
    if config['rabbitmq-host']:
        from udplog import rabbitmq
        factory = UDPLogClientFactory(
            rabbitmq.RabbitMQPublisher, dispatcher,
            vhost=config['rabbitmq-vhost'],
            exchange=config['rabbitmq-exchange'],
            queueSize=config['rabbitmq-queue-size'])
        rabbitmqClient = internet.TCPClient(config['rabbitmq-host'],
                                            config['rabbitmq-port'],
                                            factory)
        rabbitmqClient.setServiceParent(s)

    # Set up Redis client.
    if config['redis-hosts']:
        from udplog import redis
        redisService = redis.makeService(config, dispatcher)
        redisService.setServiceParent(s)

    if config['verbose']:
        UDPLogToTwistedLog(dispatcher)

    return s



# -*- test-case-name: udplog.test.test_rabbitmq -*-
#
# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
RabbitMQ support.

This provides a Twisted based RabbitMQ client for pushing logs to a RabbitMQ
exchange. A service like Logstash can subscribe to such an exchange to further
process logs.
"""

from __future__ import division, absolute_import

import copy

import simplejson

from twisted.internet import defer
from twisted.python import log
from twisted.python.filepath import FilePath

import txamqp.spec
from txamqp.client import TwistedDelegate
from txamqp.content import Content
from txamqp.protocol import AMQClient

from udplog.twisted import QueueProducer

class RabbitMQPublisher(AMQClient):
    """
    Protocol for passing UDP Log events to Logstash via RabbitMQ.

    To handle push-back from the RabbitMQ server, log events are sent
    from a local in-memory queue. If L{queueSize} is set and the queue is full,
    old items are dropped in favor of the new ones.

    @ivar queueSize: Optional maximum queue size.
    """

    def __init__(self, dispatcher, username='guest', password='guest',
                       vhost='/', exchange='logs', queueSize=None):
        self.dispatcher = dispatcher
        self.username = username
        self.password = password
        self.exchange = exchange
        self.queueSize = queueSize

        self.chan = None

        specDir = FilePath(__file__).parent()
        specFilePath = specDir.child('amqp0-9-1.extended.xml')
        spec = txamqp.spec.load(specFilePath.path)

        delegate = TwistedDelegate()
        AMQClient.__init__(self, delegate=delegate, vhost=vhost,
                                 spec=spec)


    def connectionMade(self):
        """
        Add this protocol as a consumer of log events.
        """
        AMQClient.connectionMade(self)

        def eb(failure):
            log.err(failure)
            self.transport.loseConnection()

        d = self.gotConnection()
        d.addErrback(eb)


    @defer.inlineCallbacks
    def gotConnection(self):
        yield self.authenticate(username=self.username,
                                password=self.password, mechanism='PLAIN')

        self.chan = yield self.channel(1)
        yield self.chan.channel_open()

        yield self.chan.exchange_declare(exchange="logs", type="topic",
                                         durable=True, auto_delete=False)

        self.producer = QueueProducer(callback=self.sendEvent,
                                      size=self.queueSize)
        self.transport.registerProducer(self.producer, streaming=True)
        self.producer.resumeProducing()

        self.dispatcher.register(self.producer.put)


    def connectionLost(self, reason):
        """
        Remove this protocol as a consumer of log events.
        """
        self.chan = None
        self.dispatcher.unregister(self.producer.put)
        log.err(reason, "Connection lost")
        AMQClient.connectionLost(self, reason)


    def sendEvent(self, event):
        """
        Write an event to Logstash.
        """
        if not self.chan:
            log.msg("No AMQP channel. Dropping event.")
            return

        event = copy.copy(event)

        # Pass timestamp on as a string, to work around Logstash bug 279.
        event['timestamp'] = repr(event['timestamp'])

        # Make sure isError is always a boolean
        if 'isError' in event:
            event['isError'] = bool(event['isError'])

        body = simplejson.dumps(event)
        content = Content(body)
        return self.chan.basic_publish(exchange=self.exchange,
                                       content=content)



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Redis support.

This provides a Twisted based publisher to a named list on one or more Redis
servers.
"""

from __future__ import division, absolute_import

import random

import simplejson

from twisted.application import internet, service
from twisted.python import log
from twisted.internet import defer

from txredis.client import RedisClientFactory

class NoClientError(Exception):
    """
    Raised when there are no connected clients.
    """



class RedisPublisher(service.Service):
    """
    Publisher that pushes events to a Redis list.
    """

    def __init__(self, dispatcher, client, key):
        self.dispatcher = dispatcher
        self.client = client
        self.key = key


    def startService(self):
        service.Service.startService(self)
        self.dispatcher.register(self.sendEvent)


    def stopService(self):
        self.dispatcher.unregister(self.sendEvent)
        service.Service.startService(self)


    def sendEvent(self, event):
        try:
            value = simplejson.dumps(event)
        except (TypeError, ValueError):
            log.err(None, "Could not encode event to JSON")
            return

        try:
            d = self.client.lpush(self.key, value)
        except:
            log.err()
        d.addErrback(lambda failure: failure.trap(NoClientError))
        d.addErrback(log.err)



class RedisPushMultiClient(object):
    """
    Redis push client for round-robin dispatch to multiple clients.

    This takes a list of client factories that are selected at random
    to dispatch each single push. If a client is not (yet) connected, the
    factory is taken out of the list of candidates, and re-added when
    the a new connection has been made.
    """

    def __init__(self, factories):
        """
        Initialize.

        @param factories: Client factories.
        @type factories: C{list} of L{RedisClientFactory}.
        """
        self.factories = set(factories)


    def _reconnected(self, factory):
        """
        Called when a new connection for this factory has been made.
        """
        self.factories.add(factory)


    def _disconnected(self, factory):
        """
        Called when the connection for this factory is gone.
        """
        self.factories.remove(factory)
        factory.deferred.addCallback(lambda _: self._reconnected(factory))


    def lpush(self, key, *values, **kwargs):
        """
        Add string to head of list.

        This selects a factory and attempts a push there, falling back to
        others until none are left. In that case, L{NoClientError} is fired
        from the returned deferred.

        @param key: List key
        @param values: Sequence of values to push
        @param value: For backwards compatibility, a single value.
        """
        def eb(failure):
            failure.trap(RuntimeError)
            if failure.value.args == ("Not connected",):
                self._disconnected(factory)
                return self.lpush(key, *values, **kwargs)
            else:
                return failure

        if not self.factories:
            return defer.fail(NoClientError())

        factory = random.sample(self.factories, 1)[0]
        try:
            d = factory.client.lpush(key, *values, **kwargs)
        except AttributeError:
            self._disconnected(factory)
            return self.lpush(key, *values, **kwargs)
        d.addErrback(eb)
        return d


def makeService(config, dispatcher):
    """
    Set up Redis client services.
    """
    s = service.MultiService()

    factories = []

    for host in config['redis-hosts']:
        factory = RedisClientFactory()
        factories.append(factory)
        tcpClient = internet.TCPClient(host,
                                       config['redis-port'],
                                       factory)
        tcpClient.setServiceParent(s)

    client = RedisPushMultiClient(factories)

    publisher = RedisPublisher(dispatcher, client, config['redis-key'])
    publisher.setServiceParent(s)

    return s



# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Twisted Plugins for udplog services.
"""

from __future__ import division, absolute_import

from twisted.application.service import ServiceMaker

UDPLogServer = ServiceMaker(
    "UDPLog Server",
    "udplog.tap",
    "A service that accepts structured logs via UDP and dispatches them to "
        "Scribe or RabbitMQ.",
    "udplog")



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{udplog.redis}.
"""

from __future__ import division, absolute_import

import simplejson

from twisted.application.internet import TCPClient
from twisted.internet import defer
from twisted.trial import unittest

from udplog import redis
from udplog.twisted import DispatcherFromUDPLogProtocol

class FakeRedisClient(object):

    def __init__(self):
        self.pushes = []
        self.disconnected = False


    def lpush(self, key, *values, **kwargs):
        if self.disconnected:
            return defer.fail(RuntimeError("Not connected"))
        else:
            self.pushes.append((key, values, kwargs))
            return defer.succeed(len(self.pushes))


class RedisPublisherServiceTest(unittest.TestCase):

    def setUp(self):
        self.dispatcher = DispatcherFromUDPLogProtocol()
        self.client = FakeRedisClient()
        self.publisher = redis.RedisPublisher(self.dispatcher,
                                              self.client,
                                              "test_list")


    def test_startService(self):
        """
        The publisher registers itself with the dispatcher.
        """
        event = {'message': 'test'}

        self.dispatcher.eventReceived(event)
        self.assertEqual(0, len(self.client.pushes))

        self.publisher.startService()

        self.dispatcher.eventReceived(event)
        self.assertEqual(1, len(self.client.pushes))


    def test_stopService(self):
        """
        The publisher registers itself with the dispatcher.
        """
        event = {'message': 'test'}

        self.publisher.startService()
        self.dispatcher.eventReceived(event)

        self.publisher.stopService()

        self.dispatcher.eventReceived(event)
        self.assertEqual(1, len(self.client.pushes))


    def test_sendEvent(self):
        """
        An event is pushed as a JSON string.
        """
        event = {'category': u'test',
                 'message': u'test',
                 'timestamp': 1340634165}
        self.publisher.sendEvent(event)

        output = self.client.pushes[-1]
        self.assertEqual('test_list', output[0])
        eventDict = simplejson.loads(output[1][0])
        self.assertEqual(u'test', eventDict['message'])


    def test_sendEventUnserializable(self):
        """
        An event that cannot be serialized is dropped and an error logged.
        """
        class Object(object):
            pass

        event = {'category': u'test',
                 'message': Object(),
                 'timestamp': 1340634165}
        self.publisher.sendEvent(event)

        self.assertEqual(0, len(self.client.pushes))
        self.assertEqual(1, len(self.flushLoggedErrors(TypeError)))


    def test_sendEventNoClient(self):
        """
        An event that cannot be serialized is dropped and an error logged.
        """
        event = {'category': u'test',
                 'message': u'test',
                 'timestamp': 1340634165}

        def lpush(key, *args, **kwargs):
            return defer.fail(redis.NoClientError())

        self.patch(self.client, "lpush", lpush)

        self.publisher.sendEvent(event)

        self.assertEqual(0, len(self.client.pushes))
        self.assertEqual(0, len(self.flushLoggedErrors()),
                         "Unexpected error logged")



class FakeFactory(object):

    def __init__(self, client):
        self.client = client

        self.deferred = defer.Deferred()

        if client is not None:
            self.deferred.callback(client)


    def disconnect(self):
        self.client.disconnected = True
        self.deferred = defer.Deferred()



class RedisPushMultiClientTest(unittest.TestCase):

    def test_lpush(self):
        """
        An lpush is passed on to the factory client.
        """
        client = FakeRedisClient()
        factories = [FakeFactory(client)]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        def cb(result):
            self.assertEqual(1, result)
            output = client.pushes[-1]
            self.assertEqual('test_list', output[0])
            self.assertEqual(value, output[1][0])

        d = multiClient.lpush('test_list', value)
        d.addCallback(cb)
        return d


    def test_lpushNoFactories(self):
        """
        If the list of factories is empty, NoClientError is raised.
        """
        factories = []
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        d = multiClient.lpush('test_list', value)
        self.assertFailure(d, redis.NoClientError)
        return d


    def test_lpushNoClient(self):
        """
        If a factory's client is not connected, it is removed from the pool.
        """
        factories = [FakeFactory(None)]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        d = multiClient.lpush('test_list', value)
        self.assertFailure(d, redis.NoClientError)
        return d


    def test_lpushRuntimeError(self):
        """
        If the list of factories is empty, NoClientError is raised.
        """
        client = FakeRedisClient()
        factories = [FakeFactory(client)]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        def lpush(key, *args, **kwargs):
            return defer.fail(RuntimeError("something"))

        self.patch(client, "lpush", lpush)
        d = multiClient.lpush('test_list', value)
        self.assertFailure(d, RuntimeError)
        return d


    def test_lpushMultiple(self):
        """
        Pushes are distributed over multiple clients.
        """
        client1 = FakeRedisClient()
        client2 = FakeRedisClient()
        factories = [FakeFactory(client1), FakeFactory(client2)]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        def cb(result):
            self.assertNotEqual(0, len(client1.pushes), "No pushes to client1")
            self.assertNotEqual(0, len(client2.pushes), "No pushes to client2")
            self.assertEqual(50, len(client1.pushes) + len(client2.pushes))

        d = defer.gatherResults([multiClient.lpush('test_list', value)
                                 for i in xrange(50)])
        d.addCallback(cb)
        return d


    def test_lpushMultipleOneDisconnected(self):
        """
        If a client is disconnected, its factory is removed from the pool.
        """
        client1 = FakeRedisClient()
        factory1 = FakeFactory(client1)
        client2 = FakeRedisClient()
        factory2 = FakeFactory(client2)
        factories = [factory1, factory2]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        def cb(result):
            self.assertNotIn(factory1, multiClient.factories)
            self.assertEqual(0, len(client1.pushes))
            self.assertEqual(50, len(client2.pushes))

        factory1.disconnect()
        d = defer.gatherResults([multiClient.lpush('test_list', value)
                                 for i in xrange(50)])
        d.addCallback(cb)
        return d


    def test_lpushMultipleReconnected(self):
        """
        If a factory reconnects, it is added back to the pool.
        """
        client1 = FakeRedisClient()
        factory1 = FakeFactory(client1)
        client2 = FakeRedisClient()
        factory2 = FakeFactory(client2)
        factories = [factory1, factory2]
        multiClient = redis.RedisPushMultiClient(factories)

        value = '{"message": "test"}'

        def onDisconnected(result):
            self.assertNotIn(factory1, multiClient.factories)
            self.assertEqual(0, len(client1.pushes))
            self.assertEqual(50, len(client2.pushes))

            client1.disconnected = False
            factory1.deferred.callback(client1)
            self.assertIn(factory1, multiClient.factories)

        factory1.disconnect()
        d = defer.gatherResults([multiClient.lpush('test_list', value)
                                 for i in xrange(50)])
        d.addCallback(onDisconnected)
        return d


class MakeServiceTest(unittest.TestCase):
    """
    Tests for L{redis.makeService}.
    """

    def test_services(self):
        """
        The right type and number of services are created.
        """
        config = {'redis-hosts': set(['10.0.0.2', '10.0.0.3']),
                  'redis-port': 6379,
                  'redis-key': 'udplog'}
        dispatcher = DispatcherFromUDPLogProtocol()
        multiService = redis.makeService(config, dispatcher)
        services = list(multiService)

        self.assertEqual(3, len(services))

        for service in services[:-1]:
            self.assertIsInstance(service, TCPClient)

        self.assertIsInstance(services[-1], redis.RedisPublisher)



import os
from setuptools import setup

# Make sure 'twisted' doesn't appear in top_level.txt

try:
    from setuptools.command import egg_info
    egg_info.write_toplevel_names
except (ImportError, AttributeError):
    pass
else:
    def _top_level_package(name):
        return name.split('.', 1)[0]

    def _hacked_write_toplevel_names(cmd, basename, filename):
        pkgs = dict.fromkeys(
            [_top_level_package(k)
                for k in cmd.distribution.iter_distribution_names()
                if _top_level_package(k) != "twisted"
            ]
        )
        cmd.write_file("top-level names", filename, '\n'.join(pkgs) + '\n')

    egg_info.write_toplevel_names = _hacked_write_toplevel_names

# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "udplog",
    version = "0.1.1",
    author = "Mochi Media",
    author_email = "dev@mochimedia.com",
    maintainer = "Ralph Meijer",
    maintainer_email = "ralphm@ik.nu",
    description = ("UDPLog is a system for emitting application log events"
                   " via UDP and shipping them via RabbitMQ or Scribe for "
                   "further processing."),
    license = "MIT",
    keywords = "logging twisted udp scribe",
    url = "https://github.com/mochi/udplog",
    packages=[
        'udplog',
        'udplog.test',
        'twisted.plugins',
    ],
    package_data={
        'twisted.plugins': ['twisted/plugins/server.py'],
        'udplog': ['amqp0-9-1.extended.xml'],
    },
    zip_safe=False,
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[
        'simplejson',
        'Twisted >= 13.0.0',
        'txAMQP',
        'txredis',
        'thrift',
        'scribe',
    ],
)



# -*- test-case-name: udplog.test.test_scribe -*-
#
# Copyright (c) Mochi Media, Inc.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Twisted utilities for UDP logging.

This provides a Twisted protocol implementation of the UDP Log protocol,
a handler for passing messages from the Python logging system to the Twisted
Logging system and various helpers for creating a UDP Log server.
"""

from __future__ import division, absolute_import

from collections import deque
import logging

import simplejson

from zope.interface import implements

from twisted.internet import defer
from twisted.internet import protocol
from twisted.internet.interfaces import IPushProducer
from twisted.python import log
from twisted.python import reflect
from twisted.python.failure import Failure

from udplog import udplog

class UDPLogObserver(object):
    """
    Twisted Log Observer that emits to UDP log.

    @ivar defaultCategory: Defailt log category. If there is no key C{category}
        in the C{eventDict}, use this category instead.
    """

    def __init__(self, logger, defaultCategory='udplog_unknown'):
        self.logger = logger
        self.defaultCategory = defaultCategory


    def emit(self, eventDict):
        """
        Log an event.

        This converts C{eventDict} so that it can be serialized to JSON and
        sent over UDP to the logging server.

        The key C{'time'} that is automatically provided by Twisted is renamed
        to C{'timestamp'} that is used in UDP log.

        When Twisted logs an error, the associated Failure is in the
        C{eventDict} with key C{'failure'}. For warnings, C{'warning'} holds
        the warning class and its arguments, and C{'filename'}, C{'lineno'} the
        location where the warning was reported from. See
        L{twisted.python.log.textFromEventDict} for how C{'format'} is used to
        render failures and warnings.

        See L{twisted.python.log.ILogObserver}.
        """
        eventDict = eventDict.copy()
        eventDict['message'] = log.textFromEventDict(eventDict)
        eventDict['timestamp'] = eventDict['time']

        if 'warning' in eventDict:
            # Twisted passes the warning category in 'category' and the
            # warning instance in 'warning'. Override message to only contain
            # actual warning message and put the category in 'warning'.
            eventDict['message'] = reflect.safe_str(eventDict['warning'])
            eventDict['warningCategory'] = eventDict['category']
            eventDict.setdefault('logLevel', 'WARNING')
            del eventDict['category']
            del eventDict['warning']

        if 'isError' in eventDict:
            eventDict['isError'] = bool(eventDict['isError'])

        if eventDict.get('isError', False) and 'failure' in eventDict:
            # Twisted passed the failure instance in 'failure'. Add a field
            # 'excType' containing the exception type and remove 'failure'.
            # We always want to render the traceback in a separate field, so we
            # override the actual message that textFromEventDict created for
            # us.

            udplog.augmentWithFailure(eventDict,
                                      eventDict['failure'],
                                      eventDict['why']
                                      )
            del eventDict['why']
            del eventDict['failure']

        eventDict.setdefault('logLevel', 'INFO')

        # Clean up unneeded Twisted specific keys.
        #   * time is replaced by timeformat
        #   * format, if present, is used by textFromEventDict.
        for key in ('time', 'format'):
            if key in eventDict:
                del eventDict[key]

        category = eventDict.get('category', self.defaultCategory)

        self.logger.log(category, eventDict)


    def start(self):
        """
        Start observing log events.
        """
        log.addObserver(self.emit)


    def stop(self):
        """
        Stop observing log events.
        """
        log.removeObserver(self.emit)



class TwistedLogHandler(logging.Handler):

    def __init__(self, category='python_logging', publisher=None):
        logging.Handler.__init__(self)
        self.setFormatter(logging.Formatter())

        self.category = category

        if publisher:
            self.publisher = publisher
        else:
            from twisted.python.log import theLogPublisher
            self.publisher = theLogPublisher


    def emit(self, record):
        """
        Emit a record.
        """
        try:
            eventDict = {
                    'category': self.category,
                    'logLevel': record.levelname,
                    'logName': record.name,
                    'filename': record.pathname,
                    'lineno': record.lineno,
                    'funcName': record.funcName,
                    }

            eventDict['isError'] = record.levelno >= logging.ERROR

            if isinstance(record.args, dict):
                eventDict.update(record.args)

            message = record.getMessage()

            if record.exc_info:
                exc_type, exc_value, exc_traceback = record.exc_info
                eventDict['failure'] = Failure(exc_value,
                                               exc_type,
                                               exc_traceback)
                self.publisher.msg(why=message, **eventDict)
            else:
                self.publisher.msg(message, **eventDict)

        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)



class UDPLogProtocol(protocol.DatagramProtocol):
    """
    UDP Log protocol.

    Log events are received as combination of category and a message, separated
    by a colon. This message is a dictionary encoded in JSON. Upon receiving
    an event, it is decoded and passed to L{eventReceived}.
    """

    def datagramReceived(self, datagram, addr):
        data = datagram.rstrip()

        try:
            category, event = udplog.unserialize(data)
            event['category'] = category
        except (ValueError, TypeError):
            log.err()
            return

        self.eventReceived(event)


    def eventReceived(self, event):
        """
        A log event was received.

        Override this method to process events.
        """



class DispatcherFromUDPLogProtocol(UDPLogProtocol):
    """
    Adapter from UDPLogProtocol to a consumer of log events.
    """

    def __init__(self):
        self._consumers = set()


    def register(self, consumer):
        self._consumers.add(consumer)


    def unregister(self, consumer):
        try:
            self._consumers.remove(consumer)
        except KeyError:
            pass


    def eventReceived(self, event):
        for consumer in self._consumers:
            try:
                consumer(event)
            except:
                log.err()



class QueueProducer(object):
    """
    Push producer with a queue.

    This producer accepts items with L{put} that will be delivered to
    the passed callback if not paused. If the producer is paused, the
    items get put in a queue. The queue may be limited in size, which
    causes it to only queue the last n items. Upon resuming, items from
    the queue will be passed to the callback again.

    When the deferred returned by the callback fires, the delivery of the
    next item in the queue will be scheduled.
    """

    implements(IPushProducer)

    def __init__(self, callback, size=None, clock=None):
        """
        @param callback: Callback method that gets items passed to L{put}
            whenever the producer is not paused. The callback returns
            a deferred that is waited upon before processing the next
            item.

        @param size: Optional queue size. If the queue becomes full, old
            items are dropped.

        @param clock: An object which provides
            L{twisted.internet.interfaces.IReactorTime}.
        """
        self.callback = callback
        self.paused = True
        self.waiting = None
        self.pending = deque(maxlen=size)

        if clock is None:
            from twisted.internet import reactor
            clock = reactor
        self._clock = clock

        self._call = None


    def pauseProducing(self):
        """
        Called when the transport wants to pause receiving data.
        """
        log.msg(format="Pause producing for %(callback)s.",
                       callback=self.callback)
        self.paused = True


    def resumeProducing(self):
        """
        Called when the transport is ready to resume receiving data.
        """
        log.msg(format="Resume producing for %(callback)s.",
                       callback=self.callback)
        self.paused = False
        self._processQueue()


    def stopProducing(self):
        """
        Called when the transport can no longer deliver data.
        """
        if self.waiting:
            self.waiting.cancel()
        if self._call:
            self._call.cancel()


    def put(self, obj):
        """
        Put a new item for delivery to the protocol using the callback.
        """
        if self.waiting and not self.paused:
            d = self.waiting
            self.waiting = None
            d.callback(obj)
        else:
            self.pending.append(obj)


    def _processQueue(self):
        """
        Process the next item in queue.
        """
        self._call = None

        def reschedule(_):
            self._call = self._clock.callLater(0, self._processQueue)

        def processItem(obj):
            d = self.callback(obj)
            d.addErrback(log.err)
            d.addCallback(reschedule)
            return d

        if self.pending and not self.paused:
            obj = self.pending.popleft()
            processItem(obj)
        elif not self.waiting:
            def canceller(_):
                self.waiting = None
            def trapCancelledError(failure):
                failure.trap(defer.CancelledError)
            d = defer.Deferred(canceller=canceller)
            d.addCallback(processItem)
            d.addErrback(trapCancelledError)
            d.addErrback(log.err)
            self.waiting = d
        else:
            # Reuse existing waiting deferred
            pass



class UDPLogClientFactory(protocol.ReconnectingClientFactory):
    """
    Reconnecting client factory that resets retry delay upon connecting.
    """
    maxDelay = 30

    def __init__(self, protocolClass, *args, **kwargs):
        """
        Initialize the factory.

        @param protocolClass: The class to build a protocol with.
        @type protocolClass: L{twisted.internet.interfaces.IProtocol} provider.

        @param maxDelay: Maximum number of seconds between connection
            attempts.
        @type maxDelay: C{float}.
        """
        self.protocol = protocolClass
        self.args = args
        self.kwargs = kwargs


    def buildProtocol(self, addr):
        """
        Create a protocol and reset the retry delay.
        """
        self.resetDelay()
        p = self.protocol(*self.args, **self.kwargs)
        p.factory = self
        return p



class UDPLogToTwistedLog(object):
    """
    Consumer for L{DispatcherFromUDPLogProtocol} that logs to Twisted log.
    """

    def __init__(self, dispatcher):
        dispatcher.register(self.sendEvent)


    def sendEvent(self, eventDict):
        log.msg(simplejson.dumps(eventDict, indent=4, sort_keys=True))



"""
Example of using the standard logging facility to send events to udplog.
"""

import logging
import socket
import warnings

from udplog.udplog import UDPLogger, UDPLogHandler

# Get a logger in the idiomatic way.
logger = logging.getLogger(__name__)

# Set up logging to stdout
logging.basicConfig(level=logging.DEBUG)

# Capture warnings, too.
logging.captureWarnings(True)

# Add the UDPLog handler to the root logger.
udplogger = UDPLogger(defaultFields={
                          'appname': 'example',
                          'hostname': socket.gethostname(),
                          })
root = logging.getLogger()
root.setLevel(logging.DEBUG)
root.addHandler(UDPLogHandler(udplogger, category="python_logging"))

def main():
    logger.debug("Starting!")
    logger.info("This is a simple message")
    logger.info("This is a message with %(what)s", {'what': 'variables'})

    extra_logger = logging.LoggerAdapter(logger, {'bonus': 'extra data'})
    extra_logger.info("Bonus ahead!")

    a = {}
    try:
        print a['something']
    except:
        logger.exception("Oops!")

    warnings.warn("Don't do foo, do bar instead!", stacklevel=2)

main()



"""
Example of using the Twisted logging facility to send events to udplog.
"""

import sys
import warnings

from twisted.python import log

from udplog.udplog import UDPLogger
from udplog.twisted import UDPLogObserver

# Set up logging to stdout
log.startLogging(sys.stdout)

# Set up the udplog observer
udplogger = UDPLogger()
observer = UDPLogObserver(udplogger, defaultCategory='twisted_logging')
log.addObserver(observer.emit)

def main():
    log.msg("Starting!")
    log.msg("This is a simple message")
    log.msg(format="This is a message with %(what)s", what='variables')

    a = {}
    try:
        print a['something']
    except:
        log.err(None, "Oops!")

    warnings.warn("Don't do foo, do bar instead!", stacklevel=2)

main()
